import logging

from dbus2mqtt import AppContext
from dbus2mqtt.config import FlowActionContextSetConfig
from dbus2mqtt.flow import FlowAction, FlowExecutionContext

logger = logging.getLogger(__name__)

class ContextSetAction(FlowAction):

    def __init__(self, config: FlowActionContextSetConfig, app_context: AppContext):
        self.config = config
        self.templating = app_context.templating

    async def execute(self, context: FlowExecutionContext):

        aggregated_context = context.get_aggregated_context()

        if self.config.global_context:
            context_new = await self.templating.async_render_template(self.config.global_context, dict, aggregated_context)
            logger.debug(f"Update global_context with: {context_new}")
            context.global_flows_context.update(context_new)

        if self.config.context:

            context_new = await self.templating.async_render_template(self.config.context, dict, aggregated_context)
            logger.debug(f"Update context with: {context_new}")
            context.context.update(context_new)




import os

import dotenv

from jsonargparse import ArgumentParser

from dbus2mqtt.config import Config

FILE_DIR = os.path.dirname(__file__)

def test_home_assistant_media_player_example():

    dotenv.load_dotenv(".env.example")

    parser = ArgumentParser(default_env=True, env_prefix=False)
    parser.add_class_arguments(Config)

    cfg = parser.parse_path(f"{FILE_DIR}/../../docs/examples/home_assistant_media_player.yaml")
    config = parser.instantiate_classes(cfg)

    assert config is not None

def test_linux_desktop_example():
    dotenv.load_dotenv(".env.example")

    parser = ArgumentParser(default_env=True, env_prefix=False)
    parser.add_class_arguments(Config)

    cfg = parser.parse_path(f"{FILE_DIR}/../../docs/examples/linux_desktop.yaml")
    config = parser.instantiate_classes(cfg)

    assert config is not None



import json
import re

import dbus_next.signature as dbus_signature


def _variant_serializer(obj):
    if isinstance(obj, dbus_signature.Variant):
        return obj.value
    return obj

def unwrap_dbus_object(o):
    # an easy way to get rid of dbus_next.signature.Variant types
    res = json.dumps(o, default=_variant_serializer)
    json_obj = json.loads(res)
    return json_obj

def unwrap_dbus_objects(args):
    res = [unwrap_dbus_object(o) for o in args]
    return res

def camel_to_snake(name):
    return re.sub(r'([a-z])([A-Z])', r'\1_\2', name).lower()




import dbus_next.introspection as dbus_intr
import dbus_next.signature as dbus_signature

from dbus_next.constants import ArgDirection

from tests import mocked_app_context, mocked_dbus_client


def test_signal_handler_unwrap_args():

    app_context = mocked_app_context()
    dbus_client = mocked_dbus_client(app_context)

    dbus_signal = dbus_intr.Signal("PropertiesChanged", [
        dbus_intr.Arg(name="interface_name", signature="s", direction=[ArgDirection.IN]),
        dbus_intr.Arg(name="changed_properties", signature="a{sv}", direction=[ArgDirection.IN]),
        dbus_intr.Arg(name="invalidated_properties", signature="as", direction=[ArgDirection.IN])
    ])

    dbus_signal_state = {}
    dbus_signal_state = {
        "bus_name": "org.mpris.MediaPlayer2.vlc",
        "path": "/org/mpris/MediaPlayer2",
        "interface_name": "org.freedesktop.DBus.Properties",
        "subscription_config": None, # subscription_config,
        "signal_config": None, # signal_config,
    }

    handler = dbus_client._dbus_next_signal_handler(dbus_signal, dbus_signal_state)
    assert handler is not None

    args = [
        "org.mpris.MediaPlayer2.Player",
        {
            "CanPause": dbus_signature.Variant("b", True)
        },
        []
    ]

    # Invoke with wrapped arguments
    handler(*args)

    # Check if message is published on the event_broker
    mqtt_message = app_context.event_broker.dbus_signal_queue.sync_q.get_nowait()

    # message args should be unwrapped
    assert mqtt_message is not None
    assert mqtt_message.args == ["org.mpris.MediaPlayer2.Player", {"CanPause" : True}, []]



from dbus2mqtt.config import Config
from dbus2mqtt.event_broker import EventBroker
from dbus2mqtt.template.templating import TemplateEngine


class AppContext:
    def __init__(self, config: Config, event_broker: EventBroker, templating: TemplateEngine):
        self.config = config
        self.event_broker = event_broker
        self.templating = templating




import logging

from jinja2.exceptions import TemplateError

from dbus2mqtt import AppContext
from dbus2mqtt.config import FlowActionMqttPublishConfig
from dbus2mqtt.flow import FlowAction, FlowExecutionContext
from dbus2mqtt.mqtt.mqtt_client import MqttMessage

logger = logging.getLogger(__name__)

class MqttPublishAction(FlowAction):

    def __init__(self, config: FlowActionMqttPublishConfig, app_context: AppContext):
        self.config = config
        self.event_broker = app_context.event_broker
        self.templating = app_context.templating

    async def execute(self, context: FlowExecutionContext):

        render_context = context.get_aggregated_context()

        try:
            mqtt_topic = await self.templating.async_render_template(self.config.topic, str, render_context)

            if self.config.payload_type == "text":
                res_type = str
            else:
                res_type = dict

            payload = await self.templating.async_render_template(self.config.payload_template, res_type, render_context)

        except TemplateError as e:
            logger.warning(f"Error rendering jinja template, flow: '{context.name or ''}', msg={e}, payload_template={self.config.payload_template}, render_context={render_context}", exc_info=True)
            return
        except Exception as e:
            # Dont log full exception info to avoid log spamming on dbus errors
            # due to clients disconnecting
            logger.warning(f"Error rendering jinja template, flow: '{context.name or ''}', msg={e} payload_template={self.config.payload_template}, render_context={render_context}")
            return

        logger.debug(f"public_mqtt: flow={context.name}, payload={payload}")

        await self.event_broker.publish_to_mqtt(MqttMessage(mqtt_topic, payload, payload_serialization_type=self.config.payload_type))



#!/usr/bin/env python3
import json
import subprocess
import sys

from packaging import version


def get_versions_from_git_tags() -> list[version.Version]:
    result = subprocess.run(
        ["git", "tag"],
        capture_output=True, text=True, check=True
    )
    tags = result.stdout.strip().splitlines()
    versions = [t.lstrip("v") for t in tags if t.startswith("v")]

    parsed_versions = []
    for v in versions:
        try:
            parsed = version.parse(v)
            parsed_versions.append(parsed)
        except Exception as e:
            print(f"Error parsing version {v}: {e}")

    return parsed_versions

def latest_version_by_cycle(versions: list[version.Version], cycles: list[str]) -> dict[str, version.Version]:
    res = {}
    for cycle in cycles:
        parsed_cycle = version.parse(cycle)

        for v in versions:
            # check if version is in cycle range
            # cycle can be major or major.minor or major.minor.patch
            if len(v.release) > len(parsed_cycle.release):
                # update if version is later within cycle
                if cycle not in cycles or v > parsed_cycle:
                    res[cycle] = v
    return res

def main():
    args = sys.argv[1:]
    if not args:
        print("Usage: release-versions.py 0.1 0.2 1.0 ...")
        sys.exit(1)

    versions = get_versions_from_git_tags()
    cycles = latest_version_by_cycle(versions, args)

    # Detect the overall latest version
    overall_latest = max(versions) if versions else None

    cycle_details = []
    for cycle in args:
        latest = cycles.get(cycle)
        if latest:
            cycle_details.append({
                "cycle": cycle,
                "latestVersion": latest.base_version,
                "isLatestStable": latest == overall_latest
            })

    print(json.dumps(cycle_details))

if __name__ == "__main__":
    main()



import fnmatch
import logging

from typing import Any

from dbus_next.constants import ErrorType
from dbus_next.errors import DBusError

from dbus2mqtt.dbus.dbus_client import DbusClient

logger = logging.getLogger(__name__)

class DbusContext:

    def __init__(self, dbus_client: DbusClient):
        self.dbus_client = dbus_client

    def async_dbus_list_fn(self, bus_name_pattern: str):

        res = []

        self.dbus_client.subscriptions
        for bus_name in self.dbus_client.subscriptions.keys():
            if fnmatch.fnmatchcase(bus_name, bus_name_pattern):
                res.append(bus_name)

        return res

    async def async_dbus_call_fn(self, bus_name: str, path: str, interface: str, method:str, method_args: list[Any] = []):

        if not isinstance(method_args, list):
            # Pylance will mentiod this line is unreachable. It is not as jinja2 can pass in any type
            raise ValueError("method_args must be a list")

        proxy_object = self.dbus_client.get_proxy_object(bus_name, path)
        if not proxy_object:
            raise ValueError(f"No matching subscription found for bus_name: {bus_name}, path: {path}")

        obj_interface = proxy_object.get_interface(interface)

        return await self.dbus_client.call_dbus_interface_method(obj_interface, method, method_args)

    async def async_dbus_property_get_fn(self, bus_name: str, path: str, interface: str, property:str, default_unsupported: Any = None):

        proxy_object = self.dbus_client.get_proxy_object(bus_name, path)
        if not proxy_object:
            raise ValueError(f"No matching subscription found for bus_name: {bus_name}, path: {path}")

        obj_interface = proxy_object.get_interface(interface)

        try:
            return await self.dbus_client.get_dbus_interface_property(obj_interface, property)
        except DBusError as e:
            if e.type == ErrorType.NOT_SUPPORTED.value and default_unsupported is not None:
                return default_unsupported

def jinja_custom_dbus_functions(dbus_client: DbusClient) -> dict[str, Any]:

    dbus_context = DbusContext(dbus_client)

    custom_functions: dict[str, Any] = {}
    custom_functions.update({
        "dbus_list": dbus_context.async_dbus_list_fn,
        "dbus_call": dbus_context.async_dbus_call_fn,
        "dbus_property_get": dbus_context.async_dbus_property_get_fn
    })

    return custom_functions




import asyncio
import json
import logging

from typing import Any

import paho.mqtt.client as mqtt
import yaml

from paho.mqtt.enums import CallbackAPIVersion
from paho.mqtt.subscribeoptions import SubscribeOptions

from dbus2mqtt import AppContext
from dbus2mqtt.event_broker import MqttMessage

logger = logging.getLogger(__name__)

class MqttClient:

    def __init__(self, app_context: AppContext):
        self.config = app_context.config.mqtt
        self.event_broker = app_context.event_broker

        self.client = mqtt.Client(
            protocol=mqtt.MQTTv5,
            callback_api_version=CallbackAPIVersion.VERSION2
        )

        self.client.username_pw_set(
            username=self.config.username,
            password=self.config.password.get_secret_value()
        )

        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message

    def connect(self):

        # mqtt_client.on_message = lambda client, userdata, message: asyncio.create_task(mqtt_on_message(client, userdata, message))
        self.client.connect_async(
            host=self.config.host,
            port=self.config.port
        )

    async def mqtt_publish_queue_processor_task(self):

        first_message = True

        """Continuously processes messages from the async queue."""
        while True:
            msg = await self.event_broker.mqtt_publish_queue.async_q.get()  # Wait for a message

            try:
                payload = msg.payload
                type = msg.payload_serialization_type
                if isinstance(msg.payload, dict):
                    if type == "json":
                        payload = json.dumps(msg.payload)
                    elif type == "yaml":
                        payload = yaml.dump(msg.payload)
                elif type == "text":
                    payload = str(payload)

                logger.debug(f"mqtt_publish_queue_processor_task: payload={payload}")
                self.client.publish(topic=msg.topic, payload=payload)

                if first_message:
                    logger.info(f"First message published: topic={msg.topic}, payload={payload}")
                    first_message = False

            except Exception as e:
                logger.warning(f"mqtt_publish_queue_processor_task: Exception {e}", exc_info=True)
            finally:
                self.event_broker.mqtt_publish_queue.async_q.task_done()


    async def run(self):
        """Runs the MQTT loop in a non-blocking way with asyncio."""
        self.client.loop_start()  # Runs Paho's loop in a background thread
        await asyncio.Event().wait()  # Keeps the coroutine alive

    # The callback for when the client receives a CONNACK response from the server.
    def on_connect(self, client: mqtt.Client, userdata, flags, reason_code, properties):
        if reason_code.is_failure:
            logger.warning(f"on_connect: Failed to connect: {reason_code}. Will retry connection")
        else:
            logger.info(f"on_connect: Connected to {self.config.host}:{self.config.port}")
            # Subscribing in on_connect() means that if we lose the connection and
            # reconnect then subscriptions will be renewed.
            client.subscribe("dbus2mqtt/#", options=SubscribeOptions(noLocal=True))

    def on_message(self, client: mqtt.Client, userdata: Any, msg: mqtt.MQTTMessage):

        payload = msg.payload.decode()
        if msg.retain:
            logger.info(f"on_message: skipping msg with retain=True, topic={msg.topic}, payload={payload}")
            return

        try:
            json_payload = json.loads(payload)
            logger.debug(f"on_message: msg.topic={msg.topic}, msg.payload={json.dumps(json_payload)}")
            self.event_broker.on_mqtt_receive(MqttMessage(msg.topic, json_payload))
        except json.JSONDecodeError as e:
            logger.warning(f"on_message: Unexpected payload, expecting json, topic={msg.topic}, payload={payload}, error={e}")



import pytest

from dbus2mqtt.config import (
    FlowActionContextSetConfig,
    FlowTriggerBusNameAddedConfig,
    FlowTriggerBusNameRemovedConfig,
    FlowTriggerDbusSignalConfig,
    SignalConfig,
)
from dbus2mqtt.event_broker import DbusSignalWithState
from tests import mocked_app_context, mocked_dbus_client, mocked_flow_processor


@pytest.mark.asyncio
async def test_bus_name_added_trigger():

    app_context = mocked_app_context()

    trigger_config = FlowTriggerBusNameAddedConfig()
    processor, _ = mocked_flow_processor(app_context, trigger_config, actions=[
        FlowActionContextSetConfig(
            global_context={
                "res": {
                    "bus_name": "{{ bus_name }}"
                }
            }
        )
    ])

    dbus_client = mocked_dbus_client(app_context)

    subscription_config = app_context.config.dbus.subscriptions[0]

    # trigger dbus_client and capture the triggered message
    await dbus_client._trigger_bus_name_added(subscription_config, "test-bus-name")
    trigger = app_context.event_broker.flow_trigger_queue.sync_q.get_nowait()

    # execute all flow actions
    await processor._process_flow_trigger(trigger)

    # expected context from _trigger_bus_name_added
    assert processor._global_context["res"] == {
        "bus_name": "test-bus-name",
    }

@pytest.mark.asyncio
async def test_bus_name_removed_trigger():

    app_context = mocked_app_context()

    trigger_config = FlowTriggerBusNameRemovedConfig()
    processor, _ = mocked_flow_processor(app_context, trigger_config, actions=[
        FlowActionContextSetConfig(
            global_context={
                "res": {
                    "bus_name": "{{ bus_name }}"
                }
            }
        )
    ])

    dbus_client = mocked_dbus_client(app_context)

    subscription_config = app_context.config.dbus.subscriptions[0]

    # trigger dbus_client and capture the triggered message
    await dbus_client._trigger_bus_name_removed(subscription_config, "test-bus-name")
    trigger = app_context.event_broker.flow_trigger_queue.sync_q.get_nowait()

    # execute all flow actions
    await processor._process_flow_trigger(trigger)

    # expected context from _trigger_bus_name_added
    assert processor._global_context["res"] == {
        "bus_name": "test-bus-name",
    }

@pytest.mark.asyncio
async def test_dbus_signal_trigger():

    app_context = mocked_app_context()

    trigger_config = FlowTriggerDbusSignalConfig(
        interface="test-interface-name",
        signal="TestSignal"
    )
    processor, _ = mocked_flow_processor(app_context, trigger_config, actions=[
        FlowActionContextSetConfig(
            global_context={
                "res": {
                    "bus_name": "{{ bus_name }}",
                    "path": "{{ path }}",
                    "interface": "{{ interface }}",
                    "args": "{{ args }}"
                }
            }
        )
    ])

    subscription_config = app_context.config.dbus.subscriptions[0]

    dbus_client = mocked_dbus_client(app_context)

    signal = DbusSignalWithState(
        bus_name="test-bus-name",
        path="/",
        interface_name=subscription_config.interfaces[0].interface,
        subscription_config=subscription_config,
        signal_config=SignalConfig(signal="TestSignal"),
        args=[
            "first-arg",
            "second-arg"
        ]
    )

    # trigger dbus_client and capture the triggered message
    await dbus_client._handle_on_dbus_signal(signal)
    trigger = app_context.event_broker.flow_trigger_queue.sync_q.get_nowait()

    # execute all flow actions
    await processor._process_flow_trigger(trigger)

    # validate results
    assert processor._global_context["res"] == {
        "bus_name": "test-bus-name",
        "path": "/",
        "interface": "test-interface-name",
        "args": ["first-arg", "second-arg"]
    }
